-- NamePlateGrid, by pankker / Maunotavast Zenedar-EU
-- NamePlateGrid is hereby placed in the Public Domain.

--[[
   Some technical stuff about NPG.
   The basic gist, in steps, goes something like this:
   1. Make a grid of SecureHandlerEnterLeave frames and assigns them an ID (NPG uses 1-50).

   2. When the cursor enters one such handler, we:
    * Move a nameplate matching the ID inside the handler
        (calling SetAllPoints(handler) is not enough. You have to manually attach all four courner points, i.e.:
        SetPoint("TOPLEFT", handler, "TOPLEFT"), SETPOINT("TOPRIGHT", handler, "TOPRIGHT"), ... "BOTTOMLEFT", "BOTTOMRIGHT")
        
       The IDs are assigned to the nameplates in the order they appear in. NPG also makes sure a nameplate ID does not change until 
       the nameplate has disappeared, meaning the nameplates will stay still in the grid.
       (Having the nameplates flickering around was rather irritating, and made it quite a bit harder to manage dots, and, well, anything)
       
    * Hide the EnterLeave handler so the cursor is able to point at the nameplate
    * Hide the nameplate as well, since it gets all garbagy. Well, don't hide, set the alpha to 0. Otherwise we couldn't interact with it :P
    * Enable binds

   3. Detecting when the cursor leaves a nameplate cell is a bit trickier, since the EnterLeave handler is hidden now.
    a. If we move the cursor to another cell, we'll know for sure that we are no longer in the first cell, so we will:
     * Move currently active nameplate back. 
       (Since we can't clear single anchor points, we'll do it by clearing all points and giving a small nudge to the camera with MoveViewUpStart. 
       This causes all nameplate positions to recalculate and our borrowed nameplate to return to its original place.)
     * Show EnterLeave handler on first cell.
     * Do step 2. on the new cell.

    b. To detect when the cursor leaves the NPG container entirely, we use a StateHandler. The next bit is slightly in the gray area maybe :P
       We set a StateDriver to constantly set the handler state to "1", and in the handler script we set the state to something else 
       (remembering to prevent it from recursing, of course ;P - oh, funny side note, it doesn't actually recurse forever, but about 480 times per frame IIRC.).
       This enables us to execute stuff in the restricted environment every time the StateDriver does it's thing (every 0.2 seconds).
       We'll use this to monitor when the mouse cursor is NOT over the NPG container frame, and then we'll know to disable all the stuff that need disabling.
      
   ----

   The StateHandler thingy mentioned in 3.b. is also used in monitoring the nameplates. It's not COMPLETELY necessary, but it makes things a lot easier.
   Otherwise we'd have to manage things separately in the insecure side of things and hope they match correctly when we do get an "actual" user induced oppurtunity
   to execute secure code.

--]]

local ADDON_NAME, NPG = ...

local min = _G.math.min
local wipe = _G.wipe
local tinsert = _G.tinsert
local tremove = _G.tremove
local ipairs = _G.ipairs
local pairs = _G.pairs
local WorldFrame = _G.WorldFrame
local UnitGUID = _G.UnitGUID
local SetFrameRef = NPG._addon.SetFrameRef
local tContains = NPG._addon.tContains
local print = NPG._addon.print

------------------------------------------------------------------------
-- Camera
------------------------------------------------------------------------
NPG.Camera = {}
do
   local CamFrame = CreateFrame("Frame")
   local MoveActive = nil
   
   -- Function to nudge the camera by an unnoticable amount.
   --
   -- This'll cause all nameplate positions to be recalculated, which we'll need to do when 
   -- unanchoring a nameplate from our own location to ensure that it immediately returns to its original place.
   -- (Otherwise they'd only return when the player manually turns the camera or when the unit owning the nameplate moves)
   function NPG.Camera:Nudge()
      local c = CamFrame
      c.dir = (c.dir or 1) * -1 -- change direction for each call ;P
      c.moved = nil
      MoveViewUpStart(0.001 * c.dir)
      c:SetScript("OnUpdate", function(self)
            if self.moved then
               MoveViewUpStart(0) -- use Start(0) instead of Stop(), so we don't need to wait the extra frame for future nudges.
               self:SetScript("OnUpdate",nil)
            else
               if MoveActive then
                  self.moved = 1
               else
                  -- If MoveView has been stopped previously (with MoveViewUpStop), 
                  -- we need to wait an additional frame before any movement takes place.
                  MoveActive = 1
               end
            end
      end)
   end
   
   -- keep track of MoveViewUpStop calls
   hooksecurefunc("MoveViewUpStop", function() MoveActive = nil end)
end

------------------
-- Substitute frame
------------------

do

end

------------------
-- Core
------------------
NPG.Core = {}
do
   local CoreFrame      = CreateFrame("Frame", nil, nil, "SecureHandlerAttributeTemplate")
   local ContainerFrame = CreateFrame("Frame", "NPGContainer", UIParent, "SecureHandlerBaseTemplate")
   local UpdaterFrame   = CreateFrame("Frame", nil, nil, "SecureHandlerStateTemplate")
   local PlateFrames    = {}
   local ActiveSlot = nil
   
   local NumActiveFrames = 0
   ContainerFrame:Hide()
   ContainerFrame:SetPoint("CENTER")
   
   function NPG.Core:GetActiveSlot()
      return ActiveSlot
   end
   
   local function CreatePlateFrame(id)
      local f = CreateFrame("frame", "NPGOverlay"..id, nil, "SecureHandlerEnterLeaveTemplate")
      --f.t = f:CreateTexture()
      --f.t:SetAllPoints()
      --f.t:SetTexture(1,1,0,.6)
      return f
   end
   
   local function SetupPlateFrame(plate, id)
      plate:Show()
      plate:SetParent(ContainerFrame)
      SetFrameRef(plate, "core", CoreFrame)
      plate:Execute("id = " .. id)
      CoreFrame:SetFrameRef("pbtn", plate)
      CoreFrame:Execute("plateFrames["..id.."] = self:GetFrameRef('pbtn')")
      
      plate:SetAttribute("_onenter", [[
      core:RunAttribute("_platemouseover", id)
      ]])
      
      plate:SetAttribute("_onleave", nil)
   end
   
   local function CreatePlateFrames(numFrames)
      for id = #PlateFrames + 1, numFrames do
         local f = CreatePlateFrame(id)
         PlateFrames[id] = f
         SetupPlateFrame(f, id)         
      end
   end
   
   local function SetupRefs()
      local BindsHandlerFrame = NPG.Binds:GetHandler()
      local BindsButton = NPG.Binds:GetButton()
      
      SetFrameRef(CoreFrame, "worldframe", WorldFrame)
      SetFrameRef(CoreFrame, "updater", UpdaterFrame)
      SetFrameRef(CoreFrame, "binds", BindsHandlerFrame)
      SetFrameRef(UpdaterFrame, "core", CoreFrame)
      SetFrameRef(UpdaterFrame, "container", ContainerFrame)
      SetFrameRef(UpdaterFrame, "binds", BindsHandlerFrame)
      SetFrameRef(BindsHandlerFrame, "actionbtn", BindsButton)
      SetFrameRef(BindsButton, "core", CoreFrame)
   end
   
   UpdaterFrame:SetAttribute("_onstate-update", [[
   newstate = tonumber(newstate)
   if newstate > 0 then 
      self:SetAttribute("state-update", "0")

      if newstate == 2 and not active then
         active = 1
         binds:RunAttribute("_enablebinds")
      end

      if newstate == 1 then
         core:RunAttribute("_refreshplates")
      
         if active and not container:IsUnderMouse() then
            active = nil
            binds:RunAttribute("_disablebinds")
            core:RunAttribute("_releaseactiveplate")
         end
      end
   end
   ]])
   
   do
      local core = CoreFrame
      
      core:Execute([[
      plates = newtable()
      visiblePlates = newtable()
      visiblePlates.contains = newtable()
      listedPlates = newtable()
      listedPlates.contains = newtable()
      listedPlates.size = 0
      maxListedPlates = 0
      plateIDs = newtable()
      anchors = newtable()
      plateFrames = newtable()
      tmp = newtable()
      activePlate = nil
      activeId = nil
      ]])
      
      function core:NudgeView()
         NPG.Camera:Nudge()
      end
      
      function core:UpdateFrames(...)
         NPG.Frames:Update(...)
      end
      
      function core:HideNamePlate(id)
         -- We're not "hiding", but rather setting 0 alpha on child frames of a plate.
         -- Shouldn't cause interference with other plate addons.
         NPG.Plates:HideID(id)
         NPG.Arrow:Show(id)
      end
      
      function core:ShowNamePlate(id)
         NPG.Plates:ShowID(id)
         NPG.Arrow:Hide()
      end
      
      function core:OnNewPlate(id)
         NPG.Plates:GetPlate(id)
      end
      
      function core:SlotActive(id)
         ActiveSlot = id
      end
      
      core:SetAttribute("_platemouseover", [[
      local id = ...
      self:RunAttribute("_setactiveplate", id)
      self:CallMethod("SlotActive", id)
      updater:SetAttribute("state-update", "2")
      ]])
      
      core:SetAttribute("_setactiveplate", [[
      local id = ...
      
      if not id then
         if activeId then
            id = activeId
         end
      end

      if not id then return end
      
      local plate = listedPlates[id]
      if (activePlate and activePlate ~= plate) or (activeId and activeId ~= id) then
         self:RunAttribute("_releaseactiveplate")
      end

      local pf = plateFrames[id]
      if not pf then activeId = nil return end
      
      pf:Hide()
      activeId = id
      
      if plate and activePlate ~= plate then
         local a = anchors[plate]
         a:SetPoint("TOPLEFT",pf,"TOPLEFT", -2, 2)
         a:SetPoint("TOPRIGHT",pf,"TOPRIGHT", 2, 2)
         a:SetPoint("BOTTOMLEFT",pf,"BOTTOMLEFT", -2, -2)
         a:SetPoint("BOTTOMRIGHT",pf,"BOTTOMRIGHT", 2, -2)
         self:CallMethod("HideNamePlate", plateIDs[plate])

         activePlate = plate
      end
      ]])
      
      core:SetAttribute("_releaseactiveplate", [[
      if activePlate then
         anchors[activePlate]:ClearAllPoints()
         self:CallMethod("ShowNamePlate", plateIDs[activePlate])
         activePlate = nil
         self:CallMethod("NudgeView")
      end
      
      if activeId then
         plateFrames[activeId]:Show()
         activeId = nil
         self:CallMethod("SlotActive", nil)
      end
      ]])
      
      core:SetAttribute("_gettarget", [[
      if activePlate and activePlate:IsUnderMouse() 
            and activePlate:IsVisible() then
            
         return UnitExists("mouseover") and "mouseover" or "target"
      end
      return nil
      ]])
      
      core:SetAttribute("_refreshplates", [[
      for i = #visiblePlates, 1, -1 do
         local plate = visiblePlates[i]
         if not (plate and plate:IsVisible()) then
            tremove(visiblePlates, i)
            visiblePlates.contains[plate] = nil
            
            local listId = plate and listedPlates.contains[plate]
            if listId then
               listedPlates.size = listedPlates.size - 1
               listedPlates.contains[plate] = nil
               listedPlates[listId] = nil
            end
         end
      end

      worldframe:GetChildList(wipe(tmp))
      for _, plate in ipairs(tmp) do
         local id = plateIDs[plate]
         
         if id == nil then
            local n = plate and plate.GetName and plate:GetName() or ""
            if n:sub(1,9) == "NamePlate" then
               id = tonumber(n:sub(10))
               
               local _, relFrame = plate:GetPoint(1)
               anchors[plate] = relFrame
               
               plateIDs[plate] = id
               self:CallMethod("OnNewPlate", id)
            else
               plateIDs[plate] = false
            end
         end
         
         if id then
            if plate:IsVisible() then
               if not visiblePlates.contains[plate] then
                  visiblePlates[#visiblePlates + 1] = plate
                  visiblePlates.contains[plate] = id
               end
               
               if listedPlates.size < maxListedPlates
                     and not listedPlates.contains[plate] then
                  for i = 1, maxListedPlates do
                     if not listedPlates[i] then
                        listedPlates[i] = plate
                        listedPlates.contains[plate] = i
                        listedPlates.size = listedPlates.size + 1
                        break
                     end
                  end
               end
            end
         end
      end
      
      self:RunAttribute("_updateinsecureframes")
      self:RunAttribute("_setactiveplate")
      ]])
      
      core:SetAttribute("_updateinsecureframes", [[
      wipe(tmp)
      local size = 0
      for i = 1, maxListedPlates do
         tmp[i] = plateIDs[listedPlates[i] ] or false
      end
      self:CallMethod("UpdateFrames", unpack(tmp))
      ]])
   end
   
   local function ShowFrames(numShown)
      for id, plate in ipairs(PlateFrames) do
         if id <= numShown then
            plate:Show()
         else
            plate:Hide()
         end
      end
   end
   
   function NPG.Core:CreatePlateFrames(numFrames)
      NumActiveFrames = numFrames
      CoreFrame:Execute("maxListedPlates = " .. numFrames)
      CreatePlateFrames(numFrames)
      ShowFrames(numFrames)
   end
   
   function NPG.Core:GetNumActiveFrames()
      return NumActiveFrames
   end
   
   function NPG.Core:Init()
      SetupRefs()
      NPG.Binds:Setup()
      RegisterStateDriver(UpdaterFrame, "update", "1")
      UpdaterFrame:SetAttribute("state-update", "1")
      ContainerFrame:Show()
      NPG.Camera:Nudge() -- activate MoveView.
      NPG.Frames:EnableHooks()
      NPG.GUID:EnableHooks()
   end
   
   function NPG.Core:GetPlateFrames()
      return PlateFrames
   end
   
   function NPG.Core:GetContainerFrame()
      return ContainerFrame
   end
end

NPG.Core:Init()
PASSINKKO = NPG.Frames
