------------------------------------------------------------------------
-- Frames
------------------------------------------------------------------------

local ADDON_NAME, NPG = ...

local min = _G.math.min
local wipe = _G.wipe
local tinsert = _G.tinsert
local tremove = _G.tremove
local ipairs = _G.ipairs
local pairs = _G.pairs
local WorldFrame = _G.WorldFrame
local UnitGUID = _G.UnitGUID

NPG.Frames = {}
do
   local BackgroundFrame = CreateFrame("Frame")
   local MAX_BUTTONS = 50
   local BORDERS = {
      -- 1
      {edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 8,
       insets = {top = 1/4, bottom = 1/4, left = 1/4, right = 1/4} -- not the real insets, we'll multiply these with edgeSize later.
      },
      
      -- 2
      {edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", edgeSize = 8,
       insets = {top = 5/16, bottom = 5/16, left = 5/16, right = 5/16}
      },
      
      -- 3
      {edgeFile = "Interface\\Buttons\\WHITE8X8", edgeSize = 8,
       insets = {top = 1, bottom = 1, left = 1, right = 1}
      }
   }
   
   local VisualFrames = {}
   local PlateInSlot = {}  -- slot  => plate mapping
   local SlotOfPlate = {}  -- plate => slot  mapping
   
   local function CreateVisualFrame(id)
      local f = CreateFrame("frame", "NPGVisual"..id)
      f:SetFrameStrata("BACKGROUND") -- NPG can't be placed on top of other UI elements, so let's appear bottom-most.
      f:EnableMouse(false)
      f:Hide()
      
      f.health = CreateFrame("StatusBar", nil, f)
      f.health:SetStatusBarTexture("Interface\\TARGETINGFRAME\\UI-StatusBar")
      f.health:EnableMouse(false)
      f.health.bg = f.health:CreateTexture()
      f.health.bg:SetTexture(0,0,0,1)
      f.health.bg:SetAllPoints()
      f.health:SetPoint("TOPLEFT", 0, 0)
      f.health:SetPoint("BOTTOMRIGHT", 0, 0)
      
      f.border = CreateFrame("Frame", nil, f.health)
      f.border:SetPoint("TOPLEFT")
      f.border:SetPoint("BOTTOMRIGHT")
      f.border:EnableMouse(false)
      
      f.regionparent = CreateFrame("Frame", nil, f)
      f.regionparent:SetAllPoints()
      f.regionparent:EnableMouse(false)
      
      f.showAnim = NPG.Animation:CreateAnimation(f, "sm(IN) a(-1) t(0,10) + d(0.25) a(1) t(0, -10)")
      f.hideAnim = NPG.Animation:CreateAnimation(f, "sm(OUT) d(0.25) a(-1) t(0,-10)")
      f.hideAnim:SetScript("OnFinished", function(s) f:Hide() end)
      return f
   end
   
   local function CreateVisualFrames(numFrames)
      NPG.Core:CreatePlateFrames(numFrames) -- make sure we have a matching amount of mouseover handler frames.
      local plateFrames = NPG.Core:GetPlateFrames()
      local container = NPG.Core:GetContainerFrame()
      BackgroundFrame:SetFrameStrata("BACKGROUND")
      BackgroundFrame:SetParent(container)
      BackgroundFrame:SetAllPoints()
      BackgroundFrame:SetBackdrop(BORDERS[1])
      BackgroundFrame.t = BackgroundFrame.t or BackgroundFrame:CreateTexture()
      BackgroundFrame.t:SetAllPoints()
      BackgroundFrame.t:SetTexture(0,0,0,0.8)
      
      for id = #VisualFrames + 1, numFrames do
         local f = CreateVisualFrame(id)
         VisualFrames[id] = f
         f:SetParent(container)
         f:SetAllPoints(plateFrames[id]) 
      end
   end
   
   local function HideFrames(from)
      for id = from or 1, #VisualFrames do
         VisualFrames[id]:Hide()
      end
   end
   
   function NPG.Frames:Grid(columns, rows, buttonWidth, buttonHeight, columnSpace, rowSpace)
      local numButtons = columns*rows
      if numButtons > MAX_BUTTONS or numButtons < 1 then return false end
      
      -- spaces need to be at least 1, to make sure only one mouseover handler is ever active.
      rowSpace = rowSpace <= 1 and 1 or rowSpace
      columnSpace = columnSpace <= 1 and 1 or columnSpace
      
      -- "sensible" minimum button sizes
      buttonWidth = buttonWidth < 8 and 8 or buttonWidth
      buttonHeight = buttonHeight < 8 and 8 or buttonHeight

      HideFrames(numButtons + 1) -- hide buttons that aren't part of new grid ;P
      CreateVisualFrames(numButtons)
      local plateFrames = NPG.Core:GetPlateFrames()
      
      for row = 0, rows - 1 do
         for column = 1, columns do
            local id = row * columns + column
            local frame = plateFrames[id]
            frame:SetSize(buttonWidth, buttonHeight)
            frame:ClearAllPoints()
            
            if id == 1 then
               frame:SetPoint("TOPLEFT", 1, -1) -- core container needs some breathing space!
            elseif column == 1 then
               frame:SetPoint("TOP", plateFrames[(row - 1) * columns + column], "BOTTOM", 0, -rowSpace)
            else
               frame:SetPoint("LEFT", plateFrames[id - 1], "RIGHT", columnSpace, 0)
            end
         end
      end
      
      local c = NPG.Core:GetContainerFrame()
      c:SetWidth(2 + (buttonWidth * columns) + (columnSpace * (columns - 1)))
      c:SetHeight(2 + (buttonHeight * rows) + (rowSpace * (rows - 1)))
      
      return true
   end
   
   do
      local env = {} -- layout string environment
      local UpdateEnv = {} -- environment for custom updater functions defined in the layout string
      
      env.CLASS = select(2,UnitClass("player"))
      
      local HEALTHFUNCS = { 
         function(health) return health end,
         function(health) 
            if health > 900000 then
               return ("%.1fM"):format(health/1000000)
            elseif health > 900 then
               return ("%.1fK"):format(health/1000)
            else
               return health
            end
         end
      }
      
      local healthFunc = HEALTHFUNCS[1]
      
      local Widgets = {Texts = {}, Icons = {}, Frames = {}}
      local Grid = {}
      local Border = {}
      
      --------------------
      -- Widget objects
      --------------------
      do
         local Widget = {}
         Widget.__index = Widget
         
         function Widget:new()
            local o = setmetatable({values = {}}, Widget)
            o:point("CENTER")
            o:size(nil, nil)
            o:alpha(1)
                        
            return o
         end
         
         function Widget:point(...)
            local anchor1, anchor2, ofsX, ofsY
         
            local args = select('#', ...)
            
            if args == 3 then
               anchor1, ofsX, ofsY = ...
            else
               anchor1, anchor2, ofsX, ofsY = ...
            end
            
            anchor1 = type(anchor1) == "string" and anchor1 or "CENTER"
            anchor2 = type(anchor2) == "string" and anchor2 or nil
            
            if anchor2 then
               local tmp = anchor1
               anchor1 = anchor2
               anchor2 = tmp
            else
               anchor2 = anchor1
            end
            
            if not (type(ofsX) == "number" and type(ofsY) == "number") then
               ofsX = 0
               ofsY = 0
            end
            
            self.values.anchor1 = anchor1
            self.values.anchor2 = anchor2
            self.values.ofsX = ofsX
            self.values.ofsY = ofsY
            
            return self
         end
         
         function Widget:width(width)
            if tonumber(width) then self.values.width = width end
            
            return self
         end
         
         function Widget:height(height)
            if tonumber(height) then self.values.height = height end
            
            return self
         end
         
         function Widget:size(width, height)
            self:width(width):height(height)
            
            return self
         end
         
         function Widget:color(r, g, b, a)         
            if tonumber(r) and tonumber(g) and tonumber(b) then
               self.values.r = r
               self.values.g = g
               self.values.b = b
            end
            self.values.alpha = tonumber(a) or self.values.alpha or 1
            
            return self
         end
         
         function Widget:alpha(alpha)
            if tonumber(alpha) then 
               self.values.alpha = alpha
            end
            
            return self
         end
         
         function Widget:func(func)
            if type(func) == "function" then
               self.values.func = func
            end
            
            return self
         end
         
         local Icon = {}
         Icon.__index = Icon
         setmetatable(Icon, {__index = Widget})
         
         function Icon:texture(texture)
            local type = texture:lower()
            if type == "raid" then
               self:func(function()
                  icon("raid", data.icon)
               end)
               
            elseif type:match("^aura%d") then
               local auranum = type:sub(5,6) + 0
               self:func(function()
                  local aura = data.aura[auranum]
                  if not aura then visible(false) return end
                  
                  icon("spell", aura.spellId or aura.spellName)
                  if aura.remaining == 0 then
                     visible(false)
                  elseif aura.remaining < (aura.baseDuration * 0.3) then -- aura is about to drop (or rather, can be refreshed without tick loss)
                     visible(true)
                     color(0.8, 0, 0, 1)
                  else
                     visible(true)
                     color(1, 1, 1, 1)
                  end
               end)
            
            end
            
            return self
         end
         
         function Icon:new(texture)
            local o = setmetatable(Widget.new(self), Icon)
            o:color(1, 1, 1, 1)
            
            o:texture(texture)
            return o
         end
         
         local Text = {}
         Text.__index = Text
         setmetatable(Text, {__index = Widget})
         
         function Text:new(text)
            local o = setmetatable(Widget.new(self), Text)
            o:text(text or "")
            o:color(1, 1, 1, 1)
            return o
         end
         
         function Text:text(text) -- ;P
            self.values.text = text
            return self
         end
         
         
         function env.text(text)
            local o = Text:new(text)
            tinsert(Widgets.Texts, o.values)
            return o
         end
         
         
         function env.icon(texture)
            local o = Icon:new(texture)
            tinsert(Widgets.Icons, o.values)
            return o
         end
      end
      
      
      function env.size(width, height)
         Grid.width = width
         Grid.height = height
      end
      
      function env.spacing(columnSpace, rowSpace)
         Grid.columnSpace = columnSpace
         Grid.rowSpace = rowSpace or columnSpace
      end
      
      env.space = env.spacing
      
      function env.grid(columns, rows, ...)
         Grid.columns = columns
         Grid.rows = rows
         
         if select('#',...) >= 3 then
            local width, height, columnSpace, rowSpace = ...
            env.size(width, height)
            env.spacing(columnSpace, rowSpace)
         end
      end
      
      function env.healthformat(format)
         if type(format) == "function" then
            healthFunc = format
         elseif type(format) == "number" then
            healthFunc = HEALTHFUNCS[format] or HEALTHFUNCS[1]
         else
            healthFunc = HEALTHFUNCS[1]
         end
      end
      
      function env.border(style)
         Border.style = style
      end
      
      function env.bordersize(size)
         Border.size = size
      end
      
      function env.bordercolor(r, g, b, a)
         Border.color = Border.color or {}
         Border.r = r
         Border.g = g
         Border.b = b
         Border.alpha = a
      end
      
      function env.borderalpha(alpha)
         Border.alpha = alpha
      end
      
      -- oh hey let's load the binds here. genious
      function env.binds(bindTable)
         NPG.Binds:LoadBinds(bindTable)
      end
      
      function env.auras(...)
         NPG.Auras:RegisterAuras(...)
      end
      
      local function Begin()
         wipe(Widgets.Texts)
         wipe(Widgets.Icons)
         env.binds(nil) -- clear binds just in case
         env.border(1)
         env.bordersize(8)
         env.bordercolor(1, 1, 1, 1)
         env.grid(5, 2, 64, 32, 5, 5) -- set some default values for the grid.
         env.auras() -- clear tracked auras
      end
      
      local function HideLayout(frame)
         if frame.texts then
            for id, text in ipairs(frame.texts) do
               text.func = nil
               text:Hide()
            end
         end
         
         if frame.icons then
            for id, icon in ipairs(frame.icons) do
               icon.func = nil
               icon:Hide()
            end
         end
      end
      
      local function AssignWidget(widget, frame, data)
         widget:Show()
         widget:ClearAllPoints()
         widget:SetPoint(data.anchor1, frame, data.anchor2, data.ofsX, data.ofsY)
         widget:SetAlpha(data.alpha)
         widget:SetSize(data.width or 0, data.height or 0)
         
         widget.func = data.func
         if widget.func then
            setfenv(widget.func, UpdateEnv)
         end
      end
      
      local borderTmp = {insets = {}}
      
      local function SetupBorder(frame)
         local border = BORDERS[Border.style]
         if border then
            borderTmp.edgeFile = border.edgeFile
            borderTmp.bgFile = nil
            borderTmp.tile = false
            borderTmp.tileSize = 0
            
            local size = Border.size
            borderTmp.edgeSize = size
            borderTmp.insets.top = border.insets.top * size
            borderTmp.insets.bottom = border.insets.bottom * size
            borderTmp.insets.left = border.insets.left * size
            borderTmp.insets.right = border.insets.right * size
            
            frame.border:SetBackdrop(borderTmp)
            frame.border:SetPoint("TOPLEFT", -borderTmp.insets.left, borderTmp.insets.top)
            frame.border:SetPoint("BOTTOMRIGHT", borderTmp.insets.bottom, -borderTmp.insets.bottom)
            
            local r, g, b, a = Border.r, Border.g, Border.b, Border.a
            if tonumber(r) and tonumber(g) and tonumber(b) then
               if tonumber(a) then
                  frame.border:SetBackdropBorderColor(r, g, b, a)
               else
                  frame.border:SetBackdropBorderColor(r, g, b)
               end
            else
               frame.border:SetBackdropBorderColor(1, 1, 1)
            end
         else
            frame.border:SetBackdrop(nil)
         end        
         
      end
      
      local function AssignLayout(frame)
         local frametexts = frame.texts or {}
         frame.texts = frametexts
         local frameicons = frame.icons or {}
         frame.icons = frameicons
         
         HideLayout(frame)
         SetupBorder(frame)
         
         for id, data in ipairs(Widgets.Texts) do
            local text = frametexts[id] or frame.border:CreateFontString()
                           -- parent stuff to border so they appear topmost
                           
            frametexts[id] = text
            AssignWidget(text, frame, data)
            text.text = data.text
            text:SetFont("Fonts\\FRIZQT__.TTF", 10, "THINOUTLINE")
            text:SetTextColor(data.r, data.g, data.b, data.alpha)
         end
         
         for id, data in ipairs(Widgets.Icons) do
            local icon = frameicons[id] or frame.border:CreateTexture()
            
            frameicons[id] = icon
            AssignWidget(icon, frame, data)
         end
      end
      
      local function End()
         NPG.Frames:Grid(Grid.columns, Grid.rows, Grid.width, Grid.height, Grid.columnSpace, Grid.rowSpace)
         
         for id, frame in pairs(VisualFrames) do
            AssignLayout(frame)
         end
      end
      
      function NPG.Frames:LoadLayoutString(layoutString)
         layoutFunc, errorMessage = loadstring(layoutString, "NPG Layout String")
         if errorMessage then
            error(errorMessage)
         end
         
         setfenv(layoutFunc, env)
         
         Begin()
         layoutFunc()
         End()
      end
      
      
      do -- Frame updating block
         local ICON_PATH = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_"
         local UpdateData = {}
         setmetatable(UpdateEnv, {__index = UpdateData})
         setmetatable(UpdateData, {__index = _G})
         
         --------------------------------------------
         -- Custom update function environment
         --------------------------------------------
         
         function UpdateEnv.text(text)
            if UpdateEnv.widget:GetObjectType() == "FontString" then
               UpdateEnv.widget.text = text -- we'll process this later in UpdateTexts (instead of calling SetText here)
            end
         end
         
         function UpdateEnv.visible(isVisible)
            if isVisible then
               UpdateEnv.widget:Show()
            else
               UpdateEnv.widget:Hide()
            end
         end
         
         function UpdateEnv.alpha(alpha)
            UpdateEnv.widget:SetAlpha(alpha)
         end
         
         function UpdateEnv.color(r, g, b, a)
            local widget = UpdateEnv.widget
            local type = widget:GetObjectType()
            if type == "FontString" then
               widget:SetTextColor(r, g, b, a)
            elseif type == "Texture" then
               widget:SetVertexColor(r, g, b, a)
            end            
         end
         
         function UpdateEnv.texture(texturePath)
            local widget = UpdateEnv.widget
            widget:SetTexture(texturePath)
         end
         
         function UpdateEnv.icon(type, id)
            type = type:lower()
            local texture
            
            if type == "raid" and tonumber(id) then
               texture = ICON_PATH .. (id or "")
            elseif type == "spell" then
               texture = select(3, GetSpellInfo(id))
            elseif type == "item" then
               texture = select(9, GetItemInfo(id))
            end
            UpdateEnv.texture(texture)
         end
         
         --------------------------------------------
         --------------------------------------------
         --------------------------------------------
         
         local KEYWORDS = {
            health = function() return healthFunc(UpdateData.health) end,
            healthpct = function() return ("%.1f%%"):format(100 * (UpdateData.health/UpdateData.maxhealth)) end,
            name = function() return UpdateData.name end
         }
         
         local KEYWORDFUNC = function(keyword) local f = KEYWORDS[keyword] return f and f() or "" end
      
         local function SetSlotPlate(slot, plateID)
            if PlateInSlot[slot] == plateID then return false end
         
            if plateID then
               SlotOfPlate[plateID] = slot
            else
               local currentPlate = PlateInSlot[slot]
               if currentPlate then SlotOfPlate[currentPlate] = nil end
            end
            
            PlateInSlot[slot] = plateID
            
            return true
         end
         
         local PlateAppeared = {}
         
         local function PlateJustAppeared(plateID)
            local appeared = PlateAppeared[plateID]
            PlateAppeared[plateID] = nil
            return appeared
         end
         
         local function RunFunc(widgetTbl)
            if widgetTbl then
               for i = 1, #widgetTbl do
                  local widget = widgetTbl[i]
                  local func = widget.func
                  if func then
                     UpdateEnv.widget = widget
                     func()
                  end
               end
            end
         end
         
         local function RunFuncs(frame)
            RunFunc(frame.frames)
            RunFunc(frame.texts)
            RunFunc(frame.icons)
         end
         
         local function UpdateTexts(frame)
            if frame.texts then
               for i = 1, #frame.texts do
                  local text = frame.texts[i]
                  text:SetText(text.text:gsub("%%(%w+)", KEYWORDFUNC))
               end
            end
         end
      
         local function UpdateFrame(slot, plateID)
            wipe(UpdateData)
            local parts = NPG.Plates:GetPlateData(plateID, UpdateData)
            local f = VisualFrames[slot]
            
            if not parts then
               if f then
                  f:Hide()
               end
               return
            end
            
            local hasChanged = SetSlotPlate(slot, plateID)
            local justAppeared = PlateJustAppeared(plateID)
                        
            if not f:IsVisible() then
               if f.hideAnim then
                  f.hideAnim:Stop()
               end
               f:Show()
               
               if f.showAnim then
                  f.showAnim:Play()
               end
            elseif justAppeared then 
               -- this'll pretty much only fire if the hook updates are disabled
               if f.swapAnim then
                  -- f.swapAnim:Play()
               end
            end
            
            UpdateEnv.data = UpdateData
            
            RunFuncs(f)
            UpdateTexts(f)
            f.health:SetMinMaxValues(UpdateData.minhealth, UpdateData.maxhealth)
            f.health:SetValue(UpdateData.health)
            f.health:SetStatusBarColor(parts.health:GetStatusBarColor())
         end
      
         local function HideFrame(slot)
            if not slot then return end
            
            local f = VisualFrames[slot]
            if f:IsVisible() then
               if f.hideAnim then
                  if not f.hideAnim:IsPlaying() then
                     f.hideAnim:Play()
                  end
               else
                  f:Hide()
               end
            end
            
            SetSlotPlate(slot, nil)
         end
         
         local function UpdatePlate(plateID)
            local slot = SlotOfPlate[plateID]
            if slot then UpdateFrame(slot, plateID) end
         end
      
         function NPG.Frames:Update(...)
            for slot = 1, NPG.Core:GetNumActiveFrames() do
               local plateID = select(slot, ...)
               if plateID then
                  UpdateFrame(slot, plateID)
               else
                  HideFrame(slot)
               end
            end
         end
         
         do
            local function Plate_OnShow(plateID, plate)
               PlateAppeared[plateID] = 1
            end
            
            local function Plate_OnHide(plateID, plate)
               HideFrame(SlotOfPlate[plateID])
            end
            
            local function EarlyUpdate(plateID)
               UpdatePlate(plateID)
            end
            
            local function Cast_OnShow(plateID, plate, parts)
               
            end
         
            function NPG.Frames:EnableHooks()
               NPG.Plates:AddHook("Plate_OnShow", Plate_OnShow)
               NPG.Plates:AddHook("Plate_OnHide", Plate_OnHide)
               NPG.Plates:AddHook("Health_OnValue", EarlyUpdate)
               NPG.Plates:AddHook("Cast_OnShow", Cast_OnShow)
            end
            
            function NPG.Frames:DisableHooks()
               NPG.Plates:RemoveHook("Plate_OnShow", Plate_OnShow)
               NPG.Plates:RemoveHook("Plate_OnHide", Plate_OnShow)
               NPG.Plates:RemoveHook("Health_OnValue", EarlyUpdate)
               NPG.Plates:RemoveHook("Cast_OnShow", Cast_OnShow)
               NPG.Plates:RemoveHook("Cast_OnHide", Cast_OnHide)
            end
         end
      end
   end 
end
